from . import errors
from ._wrapper import (
    Components,
    decode_to_qimage,
    decode_to_qpixmap,
    encode_qimage,
    encode_qpixmap,
)

__version__ = "0.2.5"


__all__ = [
    "__version__",
    "decode_to_qimage",
    "decode_to_qpixmap",
    "Components",
    "encode_qimage",
    "encode_qpixmap",
    "errors",
]



"""
build the cmake project, compile the code and install into local site-packages dir
"""

import subprocess
import sys
from pathlib import Path

_repo = Path(__file__).parent.parent
_site_packages = next((Path(sys.executable).parent.parent / "lib").glob("python*")) / "site-packages"


def main():
    subprocess.check_call(
        [
            "cmake",
            "-S",
            str(_repo),
            "-B",
            "cmake-build-release",
        ]
    )

    subprocess.check_call(
        [
            "cmake",
            "--build",
            "cmake-build-release",
            "-j",
            "8",
        ]
    )

    subprocess.check_call(
        [
            "cmake",
            "--install",
            "cmake-build-release",
            "--prefix",
            str(_site_packages),
        ]
    )


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.9,<3.11"
# ///

#  License agreement for matplotlib versions 1.3.0 and later
# =========================================================
#
# 1. This LICENSE AGREEMENT is between the Matplotlib Development Team
# ("MDT"), and the Individual or Organization ("Licensee") accessing and
# otherwise using matplotlib software in source or binary form and its
# associated documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, MDT
# hereby grants Licensee a nonexclusive, royalty-free, world-wide license
# to reproduce, analyze, test, perform and/or display publicly, prepare
# derivative works, distribute, and otherwise use matplotlib
# alone or in any derivative version, provided, however, that MDT's
# License Agreement and MDT's notice of copyright, i.e., "Copyright (c)
# 2012- Matplotlib Development Team; All Rights Reserved" are retained in
# matplotlib  alone or in any derivative version prepared by
# Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or
# incorporates matplotlib or any part thereof, and wants to
# make the derivative work available to others as provided herein, then
# Licensee hereby agrees to include in any such work a brief summary of
# the changes made to matplotlib .
#
# 4. MDT is making matplotlib available to Licensee on an "AS
# IS" basis.  MDT MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, MDT MAKES NO AND
# DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
# FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF MATPLOTLIB
# WILL NOT INFRINGE ANY THIRD PARTY RIGHTS.
#
# 5. MDT SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF MATPLOTLIB
#  FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR
# LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING
# MATPLOTLIB , OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF
# THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material
# breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any
# relationship of agency, partnership, or joint venture between MDT and
# Licensee.  This License Agreement does not grant permission to use MDT
# trademarks or trade name in a trademark sense to endorse or promote
# products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using matplotlib ,
# Licensee agrees to be bound by the terms and conditions of this License
# Agreement.
"""
Determine the name of the sdist and export to GitHub output named SDIST_NAME.

To run:
    $ python3 -m build --sdist
    $ ./ci/determine_sdist_name.py
"""

import os
import sys
from pathlib import Path

paths = [p.name for p in Path("dist").glob("*.tar.gz")]
if len(paths) != 1:
    sys.exit(f"Only a single sdist is supported, but found: {paths}")

print(f'Found SDist: "{paths[0]}"')
with open(os.environ["GITHUB_OUTPUT"], "a") as f:
    f.write(f"SDIST_NAME={paths[0]}\n")



from pathlib import Path

import cv2
import numpy as np
import pytest
from skimage.metrics import structural_similarity

from blurhash_pyside._qt import QImage

_tests_data_dir = Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def test_data():
    return _tests_data_dir


def _qimage_to_cv_mat(img: QImage) -> cv2.Mat:
    rgb32 = img.convertToFormat(QImage.Format.Format_RGB32)
    arr = np.array(rgb32.constBits(), copy=True).reshape(rgb32.height(), rgb32.width(), 4)
    return cv2.cvtColor(arr, cv2.COLOR_BGRA2RGB)


def assert_qimages_equal(img_a: QImage, img_b: QImage, tolerance: float = 0.001) -> None:
    view_a = _qimage_to_cv_mat(img_a)
    view_b = _qimage_to_cv_mat(img_b)

    score = structural_similarity(view_a, view_b, data_range=255, channel_axis=2)

    if 1.0 - score > tolerance:
        pytest.fail(f"ssim score failed with value: {score:8f}")



class BlurhashConversionError(Exception):
    """Generic conversion error"""


class BlurhashDecodingError(BlurhashConversionError):
    """error decoding a blurhash string into image data"""


class BlurhashEncodingError(BlurhashConversionError):
    """error encoding image data into a blurhash string"""



import pytest
from conftest import assert_qimages_equal

from blurhash_pyside import decode_to_qimage
from blurhash_pyside._qt import QImage
from blurhash_pyside.errors import BlurhashDecodingError


def test_decode__qimage(test_data):
    img = decode_to_qimage(
        "LGFO~6Yk^6#M@-5c,1Ex@@or[j6o",
        301,
        193,
    )

    assert img.constBits()
    assert img.format() == QImage.Format.Format_RGB32
    assert img.width() == 301
    assert img.height() == 193

    assert_qimages_equal(img, QImage(str(test_data / "decoded.png")))


@pytest.mark.parametrize(
    "blurhash_string",
    [
        "",
        " ",
        "ABC",
        "aghwmepjiSJDSFapispidfu",
    ],
)
def test_decode__bad_string(blurhash_string: str):
    with pytest.raises(BlurhashDecodingError):
        decode_to_qimage(
            blurhash_string,
            64,
            64,
        )


def test_decode__bad_size():
    with pytest.raises(BlurhashDecodingError):
        decode_to_qimage("LGFO~6Yk^6#M@-5c,1Ex@@or[j6o", 0, 0)



import sys

from demo import launch_demo
from PySide6.QtWidgets import QApplication


def main():
    app = QApplication()

    _ = launch_demo("PySide2 Blurhash Example")

    sys.exit(app.exec())


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest

from blurhash_pyside import Components, encode_qimage
from blurhash_pyside._qt import QImage
from blurhash_pyside.errors import BlurhashEncodingError


def test_encode__qimage(test_data: Path):
    img = QImage(str(test_data / "raw.bmp"))
    bh = encode_qimage(img, Components(4, 3))

    assert bh == "LGFO~6Yk^6#M@-5c,1Ex@@or[j6o"


def test_encode__invalid_components(test_data: Path):
    img = QImage(str(test_data / "raw.bmp"))
    with pytest.raises(BlurhashEncodingError):
        encode_qimage(img, Components(0, 0))


def test_encode__invalid_image():
    img = QImage()
    with pytest.raises(BlurhashEncodingError):
        encode_qimage(img, Components(4, 3))



from __future__ import annotations

from qtpy.QtCore import QMimeData, Qt, Signal
from qtpy.QtGui import QDragEnterEvent, QDropEvent, QFontMetrics, QImage, QPainter, QPaintEvent, QPalette, QPen, QPixmap
from qtpy.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSpinBox,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from blurhash_pyside import Components, decode_to_qpixmap, encode_qpixmap, errors


class ImgPreview(QWidget):
    """Preview an image or show placeholder text if no image"""

    def __init__(self, placeholder_text: str, parent: QWidget):
        super().__init__(parent=parent)

        self._placeholder_text = placeholder_text
        self._pix: QPixmap | None = None

    def setPixmap(self, pix: QPixmap | None):
        self._pix = pix
        self.update()

    @property
    def pixmap_sized(self):
        if not self._pix:
            return None
        return self._pix.scaled(
            self.rect().size(), Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation
        )

    def clear(self):
        self.setPixmap(None)

    def paintEvent(self, event: QPaintEvent):
        p = QPainter(self)
        p.setClipRect(self.rect())
        if self._pix:
            pix = self.pixmap_sized
            p.drawPixmap((self.rect().width() - pix.width()) // 2, 0, pix)
        else:
            p.setRenderHint(QPainter.RenderHint.TextAntialiasing)
            p.setPen(QPen(self.palette().color(QPalette.ColorRole.Text)))
            p.setFont(self.font())
            fm = QFontMetrics(self.font())
            text_rect = fm.boundingRect(self.rect(), Qt.AlignmentFlag.AlignCenter, self._placeholder_text)
            p.drawText(text_rect, self._placeholder_text)

            pen = QPen(self.palette().color(QPalette.ColorRole.Window))
            pen.setWidth(4)
            p.setPen(pen)
            p.drawRect(self.rect())
        p.end()
        event.accept()


class ComponentsEditor(QWidget):
    """edit x/y blurhash components"""

    value_changed = Signal(object)

    def __init__(self, x: int, y: int, parent=QWidget):
        super().__init__(parent=parent)

        _ly = QHBoxLayout()
        _ly.setContentsMargins(0, 0, 0, 0)
        self.setLayout(_ly)

        _ly.addWidget(QLabel("Components"), alignment=Qt.AlignmentFlag.AlignRight, stretch=4)

        self._x = QSpinBox(parent=self)
        self._x.setRange(1, 9)
        self._x.setValue(x)
        _ly.addWidget(self._x, stretch=2)

        self._y = QSpinBox(parent=self)
        self._y.setRange(1, 9)
        self._y.setValue(y)
        _ly.addWidget(self._y, stretch=2)

        # ---
        self._x.valueChanged.connect(lambda: self.value_changed.emit(self.components))
        self._y.valueChanged.connect(lambda: self.value_changed.emit(self.components))

    @property
    def components(self) -> Components:
        return Components(self._x.value(), self._y.value())


class EncodeDemo(QWidget):
    """demonstrate blurhash encoding"""

    def __init__(self, parent: QWidget):
        super().__init__(parent=parent)
        self.setAcceptDrops(True)

        _ly = QVBoxLayout()
        _ly.setSpacing(3)
        self.setLayout(_ly)

        self._orig = ImgPreview("Drag Image to Encode Here", parent=self)
        _ly.addWidget(self._orig, stretch=9)

        _ly_bhs = QHBoxLayout()
        _ly_bhs.setContentsMargins(0, 0, 0, 0)
        _ly.addLayout(_ly_bhs, stretch=1)

        self._bh_string = QLineEdit(parent=self)
        self._bh_string.setReadOnly(True)
        self._bh_string.setPlaceholderText("Encoded Blurhash String...")
        _ly_bhs.addWidget(self._bh_string)

        self._copy = QPushButton("Copy", parent=self)
        _ly_bhs.addWidget(self._copy)

        self._clear = QPushButton("Clear", parent=self)
        _ly_bhs.addWidget(self._clear)

        self._components = ComponentsEditor(4, 3, parent=self)
        _ly.addWidget(self._components, stretch=1)

        self._encoded = ImgPreview("Preview", parent=self)
        _ly.addWidget(self._encoded, stretch=9)

        # ---
        self._copy.clicked.connect(lambda: QApplication.clipboard().setText(self._bh_string.text()))

        self._clear.clicked.connect(lambda: self._orig.clear())
        self._clear.clicked.connect(lambda: self._encoded.clear())
        self._clear.clicked.connect(lambda: self._bh_string.clear())

        self._components.value_changed.connect(lambda: self._update_blurhash())

    def dragEnterEvent(self, event: QDragEnterEvent):
        img = self._img_from_mimeData(event.mimeData())
        if img:
            event.acceptProposedAction()
        else:
            event.ignore()

    def dropEvent(self, event: QDropEvent):
        img = self._img_from_mimeData(event.mimeData())
        if img:
            self._orig.setPixmap(QPixmap.fromImage(img))
            self._update_blurhash()
            event.accept()
        else:
            event.ignore()

    def _update_blurhash(self):
        if not self._orig.pixmap_sized:
            return

        str_val = encode_qpixmap(self._orig.pixmap_sized, self._components.components)
        self._bh_string.setText(str_val)
        self._encoded.setPixmap(
            decode_to_qpixmap(str_val, self._orig.pixmap_sized.width(), self._orig.pixmap_sized.height())
        )

    def _img_from_mimeData(self, mime_data: QMimeData) -> QImage | None:
        try:
            img = QImage(mime_data.urls()[0].toLocalFile())
            if img.constBits():
                return img
        except:
            pass
        return None

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._update_blurhash()


class DecodeDemo(QWidget):
    """demonstrate blurhash decoding"""

    def __init__(self, parent: QWidget):
        super().__init__(parent=parent)

        _ly = QVBoxLayout()
        _ly.setSpacing(3)
        self.setLayout(_ly)

        _ly_bhs = QHBoxLayout()
        _ly_bhs.setContentsMargins(0, 0, 0, 0)
        _ly.addLayout(_ly_bhs, stretch=1)

        self._bh_string = QLineEdit(parent=self)
        self._bh_string.setPlaceholderText("Enter Encoded Blurhash String Here")
        _ly_bhs.addWidget(self._bh_string)

        self._clear = QPushButton("Clear", parent=self)
        _ly_bhs.addWidget(self._clear)

        self._decoded = ImgPreview("Decoded", parent=self)
        _ly.addWidget(self._decoded, stretch=9)

        # ----
        self._bh_string.textChanged.connect(lambda: self._update_blurhash())

        self._clear.clicked.connect(lambda: self._bh_string.clear())

    def _update_blurhash(self):
        if not self._bh_string.text():
            self._decoded.clear()
            return

        try:
            decoded = decode_to_qpixmap(self._bh_string.text(), self._decoded.width(), self._decoded.height())
            if decoded:
                self._decoded.setPixmap(decoded)
        except errors.BlurhashDecodingError:
            pix = QPixmap(32, 32)
            pix.fill("red")
            # TODO: better error pixmap? draw X or msg text?
            self._decoded.setPixmap(pix)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._update_blurhash()


class BlurhashDemo(QWidget):
    """demo for blurhash encoding and decoding in 2 tabs"""

    def __init__(self, parent: QWidget):
        super().__init__(parent=parent)
        self.setMinimumSize(480, 480)

        _ly = QVBoxLayout()
        self.setLayout(_ly)

        self._tabs = QTabWidget(parent=self)
        _ly.addWidget(self._tabs)

        self._encode = EncodeDemo(parent=self)
        self._tabs.addTab(self._encode, "Encode Demo")

        self._decode = DecodeDemo(parent=self)
        self._tabs.addTab(self._decode, "Decode Demo")



_failed_pyside6 = False
_failed_pyside2 = False

try:
    from PySide6.QtCore import QSize, Qt
    from PySide6.QtGui import QImage, QPixmap
except ImportError:
    _failed_pyside6 = True

try:
    from PySide2.QtCore import QSize, Qt
    from PySide2.QtGui import QImage, QPixmap
except ImportError:
    _failed_pyside2 = True

if _failed_pyside6 and _failed_pyside2:
    raise ImportError("Could not import PySide6 or PySide2")

__all__ = [
    "QSize",
    "Qt",
    "QImage",
    "QPixmap",
]



__version__: str

def decode(
    blurhash: str,
    width: int,
    height: int,
) -> list[int]: ...
def encode(
    image: list[int],
    width: int,
    height: int,
    x: int,
    y: int,
) -> str: ...



from qtpy.QtWidgets import QMainWindow

from .blurhash_widget import BlurhashDemo


def launch_demo(title: str = "Blurhash Demo") -> QMainWindow:
    main_window = QMainWindow()
    main_window.setWindowTitle(title)
    main_window.setCentralWidget(BlurhashDemo(parent=main_window))
    main_window.show()
    return main_window



from typing import NamedTuple

from ._core import decode, encode
from ._qt import QImage, QPixmap, QSize, Qt
from .errors import BlurhashDecodingError, BlurhashEncodingError


def decode_to_qimage(blurhash: str, width: int, height: int) -> QImage:
    """
    Decode a Blurhash string to a QImage.
    """
    try:
        data = decode(blurhash.strip(), width, height)
    except Exception as e:
        raise BlurhashDecodingError(str(e)) from e

    if len(data) < height * width * 4:
        raise BlurhashDecodingError("decoded data size did not match expected height/width")

    return QImage(bytes(data), width, height, width * 4, QImage.Format.Format_RGB32).rgbSwapped()


def decode_to_qpixmap(blurhash: str, width: int, height: int) -> QPixmap:
    """
    Decode a Blurhash string to a QPixmap.
    """
    image = decode_to_qimage(blurhash, width, height)
    return QPixmap.fromImage(image)


class Components(NamedTuple):
    x: int
    y: int

    def valid(self) -> bool:
        return 9 >= self.x >= 1 and 9 >= self.y >= 1


def encode_qimage(image: QImage, components: Components, downsample: int = 1) -> str:
    """Encode a QImage to a Blurhash string"""

    if not components.valid():
        raise BlurhashEncodingError(f"Components: {components} invalid")

    image = image.convertToFormat(QImage.Format.Format_RGB32).rgbSwapped()

    if downsample > 1:
        image = image.scaled(
            QSize(downsample, downsample),
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.FastTransformation,
        )

    if not image.constBits():
        raise BlurhashEncodingError("QImage to encode seems invalid")

    bpp = 4
    size = image.width() * image.height() * bpp
    data = list(image.constBits())

    ordered = []
    try:
        for i in range(0, size, bpp):
            ordered.extend(data[i : i + bpp][0:3])
    except Exception as e:
        raise BlurhashEncodingError("problem organizing raw image data") from e

    try:
        bh_str = encode(
            ordered,
            image.width(),
            image.height(),
            components.x,
            components.y,
        )
    except Exception as e:
        raise BlurhashEncodingError(str(e)) from e

    if not bh_str:
        raise BlurhashEncodingError("Blurhash result was empty")

    return bh_str


def encode_qpixmap(pixmap: QPixmap, components: Components, downsample_factor: int = 1) -> str:
    """Encode a QPixmap to a Blurhash string"""
    return encode_qimage(pixmap.toImage(), components, downsample_factor)
